package ru.dbotthepony.mc.otm.menu;

import net.minecraft.client.Minecraft;
import net.minecraft.network.FriendlyByteBuf;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.SimpleContainer;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.ClickAction;
import net.minecraft.world.inventory.ClickType;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.fmllegacy.network.NetworkEvent;
import net.minecraftforge.fmllegacy.network.PacketDistributor;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.block.entity.BlockEntityDriveViewer;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.capability.drive.IItemViewListener;
import ru.dbotthepony.mc.otm.capability.drive.IMatteryDrive;
import ru.dbotthepony.mc.otm.item.ItemPortableCondensationDrive;
import ru.dbotthepony.mc.otm.menu.slot.MatterySlot;
import ru.dbotthepony.mc.otm.network.MatteryNetworking;
import ru.dbotthepony.mc.otm.network.SetCarriedPacket;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;
import java.util.function.Supplier;

public class DriveViewerMenu extends PoweredMatteryMenu {
	public NetworkedItemView view;
	public MatterySlot drive_slot;
	protected IMatteryDrive last_drive;

	public DriveViewerMenu(int containerID, Inventory inventory) {
		this(containerID, inventory, null);
	}

	public DriveViewerMenu(int containerID, Inventory inventory, @Nullable BlockEntityDriveViewer tile) {
		super(Registry.Menus.DRIVE_VIEWER, containerID, inventory, tile);

		var container = tile != null ? tile.drive_slot : new SimpleContainer(1);

		drive_slot = new MatterySlot(container, 0) {
			@Override
			public boolean mayPlace(ItemStack stack) {
				return stack.getCapability(MatteryCapability.DRIVE).isPresent();
			}
		};

		addSlot(drive_slot);
		view = new NetworkedItemView();

		addBatterySlot();
		addInventorySlots();
	}

	@Override
	public void broadcastChanges() {
		super.broadcastChanges();

		if (tile != null) {
			var get = ((BlockEntityDriveViewer) tile).drive_slot.getItem(0);

			if (get.isEmpty()) {
				if (last_drive != null) {
					last_drive.removeListener(view);
					view.clear();
				}

				last_drive = null;
			} else {
				var get_cap = get.getCapability(MatteryCapability.DRIVE).resolve();

				if (get_cap.isEmpty()) {
					if (last_drive != null) {
						last_drive.removeListener(view);
						view.clear();
					}

					last_drive = null;
				} else {
					if (last_drive != get_cap.get()) {
						if (last_drive != null)
							last_drive.removeListener(view);

						view.clear();
						get_cap.get().addListener(view);
					}

					last_drive = get_cap.get();
				}
			}

			view.network();
		}
	}

	@Override
	protected int getWorkingSlotStart() {
		return 0;
	}

	@Override
	protected int getWorkingSlotEnd() {
		return 2;
	}

	@Override
	public void removed(Player p_38940_) {
		super.removed(p_38940_);

		if (last_drive != null) {
			last_drive.removeListener(view);
			view.clear();
		}
	}

	/**
	 * Creates a virtual, slotless container for Player's interaction with.
	 */
	public class NetworkedItemView implements IItemViewListener {
		public record NetworkedItem(int id, ItemStack stack, UUID id_upstream) {
			public NetworkedItem(int id, ItemStack stack) {
				this(id, stack, null);
			}
		}

		protected int next_stack_id = 0;

		protected final HashMap<Integer, NetworkedItem> state = new HashMap<>();
		protected final HashMap<UUID, NetworkedItem> upstream_state = new HashMap<>();
		protected final ArrayList<Object> backlog = new ArrayList<>();

		private List<NetworkedItem> view_cache;

		public void clearCache() {
			view_cache = null;
		}

		public List<NetworkedItem> getItems() {
			if (view_cache != null)
				return view_cache;

			return view_cache = List.copyOf(state.values());
		}

		@Override
		public void addViewItem(ItemStack stack, UUID id_upstream) {
			if (upstream_state.containsKey(id_upstream))
				throw new IllegalStateException("Already tracking ItemStack with upstream id " + id_upstream + "!");

			var state = new NetworkedItem(next_stack_id++, stack.copy(), id_upstream);
			this.state.put(state.id, state);
			this.upstream_state.put(id_upstream, state);

			if (tile != null) {
				backlog.add(new StackAddPacket(containerId, state.id, stack));
			}

			clearCache();
		}

		@Override
		public void changeViewItem(UUID id_upstream, int new_count) {
			var get = upstream_state.get(id_upstream);

			if (get == null)
				throw new IllegalStateException("Unknown ItemStack with upstream id " + id_upstream + "!");

			get.stack.setCount(new_count);

			if (tile != null) {
				backlog.add(new StackChangePacket(containerId, get.id, new_count));
			}

			clearCache();
		}

		@Override
		public void removeViewItem(UUID id_upstream) {
			var get = upstream_state.get(id_upstream);

			if (get == null)
				throw new IllegalStateException("Unknown ItemStack with upstream id " + id_upstream + "!");

			upstream_state.remove(id_upstream);
			state.remove(get.id);

			if (tile != null) {
				backlog.add(new StackRemovePacket(containerId, get.id));
			}

			clearCache();
		}

		public void clear() {
			clearCache();
			upstream_state.clear();
			state.clear();

			if (tile != null) {
				backlog.clear();
				backlog.add(new ClearPacket(containerId));
			}
		}

		public void network() {
			if (tile == null)
				throw new IllegalStateException("Not a server");

			var consumer = PacketDistributor.PLAYER.with(() -> (ServerPlayer) ply);

			for (var packet : backlog) {
				MatteryNetworking.CHANNEL.send(consumer, packet);
			}

			backlog.clear();
		}
	}

	@Override
	public ItemStack quickMoveStack(Player ply, int slot_index) {
		var slot = slots.get(slot_index);
		var item = slot.getItem();

		if (item.isEmpty() || item.getCapability(MatteryCapability.DRIVE).isPresent() || item.getCapability(CapabilityEnergy.ENERGY).isPresent())
			return super.quickMoveStack(ply, slot_index);

		if (last_drive == null)
			return ItemStack.EMPTY;

		var remaining = last_drive.insertItem(item, false);

		if (remaining.getCount() == item.getCount())
			return ItemStack.EMPTY;

		if (remaining.isEmpty()) {
			var copy = item.copy();
			slot.set(ItemStack.EMPTY);
			return copy;
		}

		var copy = item.copy();
		item.setCount(item.getCount() - remaining.getCount());
		slot.setChanged();

		return copy;
	}

	public record InteractPacket(int id, int stack_id, ClickType click, ClickAction action) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
			buffer.writeInt(stack_id);
			buffer.writeEnum(click);
			buffer.writeEnum(action);
		}

		public static InteractPacket read(FriendlyByteBuf buffer) {
			var id = buffer.readInt();
			var stack = buffer.readInt();
			var click = buffer.readEnum(ClickType.class);
			var action = buffer.readEnum(ClickAction.class);
			return new InteractPacket(id, stack, click, action);
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var ply = context.get().getSender();

				if (ply.containerMenu instanceof DriveViewerMenu menu) {
					var view = menu.containerId == id ? menu.view : null;

					if (view == null) {
						return;
					}

					ply.resetLastActionTime();

					if (menu.last_drive == null)
						return;

					if (click == ClickType.QUICK_MOVE && stack_id > -1) {
						var get_state = view.state.get(stack_id);

						if (get_state == null) {
							return;
						}

						var extracted = menu.last_drive.extractItem(get_state.id_upstream, action == ClickAction.PRIMARY ? get_state.stack.getMaxStackSize() : Math.max(1, get_state.stack.getMaxStackSize() / 2), true);

						if (!extracted.isEmpty()) {
							var move = menu.quickMoveToInventory(extracted, false);

							if (move.remaining().getCount() != extracted.getCount()) {
								menu.last_drive.extractItem(get_state.id_upstream, extracted.getCount() - move.remaining().getCount(), false);
							}
						}

						return;
					}

					if (click == ClickType.CLONE) {
						if (stack_id < 0 || !ply.getAbilities().instabuild)
							return;

						var get_state = view.state.get(stack_id);

						if (get_state == null) {
							return;
						}

						var copy = get_state.stack.copy();
						copy.setCount(Math.min(copy.getCount(), copy.getMaxStackSize()));

						menu.setCarried(copy);
						MatteryNetworking.send(ply, new SetCarriedPacket(menu.getCarried()));
						menu.setRemoteCarried(menu.getCarried().copy());
						return;
					}

					if (!menu.getCarried().isEmpty() && click != ClickType.QUICK_MOVE) {
						// try to put
						if (action == ClickAction.PRIMARY) {
							menu.setCarried(menu.last_drive.insertItem(menu.getCarried(), false));
							MatteryNetworking.send(ply, new SetCarriedPacket(menu.getCarried()));
							menu.setRemoteCarried(menu.getCarried().copy());
						} else {
							var copy = menu.getCarried().copy();
							copy.setCount(1);

							if (menu.last_drive.insertItem(copy, false).isEmpty()) {
								menu.getCarried().shrink(1);
								MatteryNetworking.send(ply, new SetCarriedPacket(menu.getCarried()));
								menu.setRemoteCarried(menu.getCarried().copy());
							}
						}
					} else if (stack_id > -1) {
						var get_state = view.state.get(stack_id);

						if (get_state == null) {
							return;
						}

						menu.setCarried(menu.last_drive.extractItem(get_state.id_upstream, action == ClickAction.PRIMARY ? get_state.stack.getMaxStackSize() : Math.max(1, get_state.stack.getMaxStackSize() / 2), false));
						MatteryNetworking.send(ply, new SetCarriedPacket(menu.getCarried()));
						menu.setRemoteCarried(menu.getCarried().copy());
					}
				}
			});
		}
	}

	public record ClearPacket(int id) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
		}

		public static ClearPacket read(FriendlyByteBuf buffer) {
			return new ClearPacket(buffer.readInt());
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var get = Minecraft.getInstance().player.containerMenu instanceof DriveViewerMenu menu && menu.containerId == id ? menu.view : null;

				if (get == null) {
					throw new IllegalStateException("No such item tracker with id " + id);
				}

				get.clear();
			});
		}
	}

	public record StackAddPacket(int id, int stack_id, ItemStack stack) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
			buffer.writeInt(stack_id);
			buffer.writeRegistryId(stack.getItem());
			buffer.writeInt(stack.getCount());
			buffer.writeBoolean(stack.getTag() != null);

			if (stack.getTag() != null)
				buffer.writeNbt(stack.getShareTag());
		}

		public static StackAddPacket read(FriendlyByteBuf buffer) {
			var id = buffer.readInt();
			var stack = buffer.readInt();
			var item = buffer.readRegistryIdSafe(Item.class);
			var count = buffer.readInt();

			var state = new ItemStack(item, count);

			if (buffer.readBoolean())
				state.readShareTag(buffer.readNbt());

			return new StackAddPacket(id, stack, state);
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var get = Minecraft.getInstance().player.containerMenu instanceof DriveViewerMenu menu && menu.containerId == id ? menu.view : null;

				if (get == null) {
					throw new IllegalStateException("No such item tracker with id " + id);
				}

				if (get.state.containsKey(stack_id)) {
					throw new IllegalStateException("Item tracker " + id + " already has stack with id of " + stack_id);
				}

				get.state.put(stack_id, new NetworkedItemView.NetworkedItem(stack_id, stack));
				get.clearCache();
			});
		}
	}

	public record StackChangePacket(int id, int stack_id, int new_count) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
			buffer.writeInt(stack_id);
			buffer.writeInt(new_count);
		}

		public static StackChangePacket read(FriendlyByteBuf buffer) {
			var id = buffer.readInt();
			var stack_id = buffer.readInt();
			var new_count = buffer.readInt();

			return new StackChangePacket(id, stack_id, new_count);
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var get = Minecraft.getInstance().player.containerMenu instanceof DriveViewerMenu menu && menu.containerId == id ? menu.view : null;

				if (get == null) {
					throw new IllegalStateException("No such item tracker with id " + id);
				}

				var get_state = get.state.get(stack_id);

				if (get_state == null) {
					throw new IllegalStateException("Item tracker " + id + " has no stack with id of " + stack_id);
				}

				get_state.stack.setCount(new_count);
				get.clearCache();
			});
		}
	}

	public record StackRemovePacket(int id, int stack_id) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
			buffer.writeInt(stack_id);
		}

		public static StackRemovePacket read(FriendlyByteBuf buffer) {
			var id = buffer.readInt();
			var stack_id = buffer.readInt();

			return new StackRemovePacket(id, stack_id);
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var get = Minecraft.getInstance().player.containerMenu instanceof DriveViewerMenu menu && menu.containerId == id ? menu.view : null;

				if (get == null) {
					throw new IllegalStateException("No such item tracker with id " + id);
				}

				var get_state = get.state.remove(stack_id);

				if (get_state == null) {
					throw new IllegalStateException("Item tracker " + id + " has no stack with id of " + stack_id);
				}

				get.clearCache();
			});
		}
	}

	public enum FilterSwitch {
		MATCH_NBT,
		MATCH_TAG,
		BLACKLIST
	}

	public ItemPortableCondensationDrive.FilterSettings getFilter() {
		if (drive_slot.getItem().isEmpty())
			return null;

		if (drive_slot.getItem().getItem() instanceof ItemPortableCondensationDrive item)
			return item.getFilterSettings(drive_slot.getItem());

		return null;
	}

	public record FilterSetPacket(int id, int slot, ItemStack value) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
			buffer.writeInt(slot);
			buffer.writeItem(value);
		}

		public static FilterSetPacket read(FriendlyByteBuf buffer) {
			return new FilterSetPacket(buffer.readInt(), buffer.readInt(), buffer.readItem());
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);

			if (slot < 0 || slot >= ItemPortableCondensationDrive.FilterSettings.MAX_FILTERS)
				return;

			if (value.getCount() > 1)
				value.setCount(1);

			context.get().enqueueWork(() -> {
				var ply = context.get().getSender();

				if (ply.containerMenu instanceof DriveViewerMenu menu && menu.containerId == id) {
					if (menu.drive_slot.getItem().isEmpty())
						return;

					if (menu.drive_slot.getItem().getItem() instanceof ItemPortableCondensationDrive drive) {
						var filter = drive.getFilterSettings(menu.drive_slot.getItem());
						filter.items[slot] = value;
						filter.serializeNBT(menu.drive_slot.getItem());
					}
				}
			});
		}
	}

	public record FilterSwitchPacket(int id, FilterSwitch type, boolean value) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
			buffer.writeEnum(type);
			buffer.writeBoolean(value);
		}

		public static FilterSwitchPacket read(FriendlyByteBuf buffer) {
			return new FilterSwitchPacket(buffer.readInt(), buffer.readEnum(FilterSwitch.class), buffer.readBoolean());
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var ply = context.get().getSender();

				if (ply.containerMenu instanceof DriveViewerMenu menu && menu.containerId == id) {
					var settings = menu.getFilter();

					if (settings == null)
						return;

					switch (type) {
						case MATCH_NBT -> {
							settings.match_nbt = value;
						}

						case MATCH_TAG -> {
							settings.match_tag = value;
						}

						case BLACKLIST -> {
							settings.blacklist = value;
						}
					}

					settings.serializeNBT(menu.drive_slot.getItem());
				}
			});
		}
	}
}
