package ru.dbotthepony.mc.otm.menu.widget;

import com.mojang.blaze3d.systems.RenderSystem;
import com.mojang.blaze3d.vertex.PoseStack;
import net.minecraft.ChatFormatting;
import net.minecraft.client.renderer.GameRenderer;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.inventory.ContainerData;
import ru.dbotthepony.mc.otm.menu.MatteryMenu;
import ru.dbotthepony.mc.otm.screen.MatteryScreen;

import javax.annotation.Nullable;
import java.util.List;
import java.util.function.Supplier;

public class ProgressGaugeWidget extends GaugeWidget {
	protected final Supplier<Float> progress_supplier;
	protected final Supplier<Boolean> is_stuck_supplier;

	protected final ContainerData progress_container = new ContainerData() {
		private int progress;
		private int is_stuck;

		@Override
		public int get(int i) {
			return i == 0 ? progress : is_stuck;
		}

		@Override
		public void set(int i, int i1) {
			if (i == 0) {
				progress = i1;
			} else {
				is_stuck = i1;
			}
		}

		@Override
		public int getCount() {
			return 2;
		}
	};

	public ProgressGaugeWidget(MatteryMenu menu, int x, int y, Supplier<Float> progress, Supplier<Boolean> is_stuck_supplier) {
		super(menu, x, y);
		fill_order = GaugeDirection.LEFT_TO_RIGHT;
		addDataSlots(progress_container);
		progress_supplier = progress;
		this.is_stuck_supplier = is_stuck_supplier;
	}

	public ProgressGaugeWidget(MatteryMenu menu, int x, int y) {
		this(menu, x, y, null, null);
	}

	public ProgressGaugeWidget(MatteryMenu menu) {
		this(menu, 0, 0, null, null);
	}

	public ProgressGaugeWidget(MatteryMenu menu, Supplier<Float> progress) {
		this(menu, 0, 0, progress, null);
	}

	public ProgressGaugeWidget(MatteryMenu menu, int x, int y, Supplier<Float> progress) {
		this(menu, x, y, progress, null);
	}

	@Override
	public int getImageWidth() {
		return 22;
	}

	@Override
	public int getImageHeight() {
		return 16;
	}

	@Override
	public int getImageX() {
		return 0;
	}

	@Override
	public int getImageY() {
		return 132;
	}

	@Override
	public void updateServer() {
		if (progress_supplier != null)
			progress_container.set(0, (int) (progress_supplier.get() * 10000f));

		if (is_stuck_supplier != null)
			progress_container.set(1, is_stuck_supplier.get() ? 1 : 0);
	}

	@Override
	protected void setupGaugeDrawing() {
		super.setupGaugeDrawing();

		if (progress_container.get(1) != 0 && System.currentTimeMillis() % 1500L > 750L) {
			RenderSystem.setShaderColor(0.8F, 0.4F, 0.4F, 1.0F);
		}
	}

	@Nullable
	@Override
	public List<Component> getTooltip() {
		List<Component> text;

		if (progress_container.get(1) != 0) {
			text = List.of(
				new TranslatableComponent("otm.gui.progress_widget", String.format("%.2f", progress_container.get(0) / 100f)),
				new TranslatableComponent("otm.gui.progress_widget_stuck").withStyle(ChatFormatting.DARK_RED)
			);
		} else {
			text = List.of(
				new TranslatableComponent("otm.gui.progress_widget", String.format("%.2f", progress_container.get(0) / 100f))
			);
		}

		return text;
	}

	@Override
	public float getLevel() {
		return progress_container.get(0) / 10000f;
	}
}
