package ru.dbotthepony.mc.otm.network;

import net.minecraft.client.Minecraft;
import net.minecraft.network.FriendlyByteBuf;
import net.minecraftforge.api.distmarker.Dist;
import net.minecraftforge.fml.DistExecutor;
import net.minecraftforge.fmllegacy.network.NetworkEvent;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.PatternState;
import ru.dbotthepony.mc.otm.menu.MatterPanelMenu;

import javax.annotation.ParametersAreNonnullByDefault;
import java.util.Objects;
import java.util.function.Supplier;

@ParametersAreNonnullByDefault
public record PatternGridPacket(boolean action, PatternState ...state) {
	public void write(FriendlyByteBuf buffer) {
		buffer.writeBoolean(action);
		buffer.writeInt(state.length);

		for (var state1 : state)
			state1.write(buffer);
	}

	public void play(Supplier<NetworkEvent.Context> context) {
		context.get().setPacketHandled(true);
		context.get().enqueueWork(() -> DistExecutor.unsafeRunWhenOn(Dist.CLIENT, () -> this::playClient));
	}

	private void playClient() {
		if (Minecraft.getInstance().player.containerMenu instanceof MatterPanelMenu menu) {
			if (action)
				menu.networkPatternsUpdated(state);
			else
				menu.networkPatternsRemoved(state);

			return;
		}

		OverdriveThatMatters.LOGGER.error("Receive pattern state list, but no valid container is open / container is not MatterPanelMenu!");
	}

	public static PatternGridPacket read(FriendlyByteBuf buffer) {
		boolean action = buffer.readBoolean();
		int amount = buffer.readInt();
		var list = new PatternState[amount];

		for (int i = 0; i < amount; i++)
			list[i] = Objects.requireNonNull(PatternState.read(buffer));

		return new PatternGridPacket(action, list);
	}
}
